==> ibus-kmfl-1.0.2.tar.gz: ibus-kmfl-1.0.2/src/engine.c <==
/* vim:set et sts=4: */

#include <ibus.h>
#include <string.h>
#include <stdio.h>
#include "kmflutil.h"
#include "engine.h"

typedef struct _IBusKMFLEngine IBusKMFLEngine;
typedef struct _IBusKMFLEngineClass IBusKMFLEngineClass;

struct _IBusKMFLEngine {
	IBusEngine parent;

    /* members */
    KInputContext *context;
    IBusLookupTable *table;
    IBusProperty    *status_prop;
    IBusPropList    *prop_list;
};

struct _IBusKMFLEngineClass {
	IBusEngineClass parent;
};

/* functions prototype */
static void	ibus_kmfl_engine_class_init	    (IBusKMFLEngineClass    *klass);
static void	ibus_kmfl_engine_init		    (IBusKMFLEngine		    *kmfl);
static GObject*
            ibus_kmfl_engine_constructor    (GType                   type,
                                             guint                   n_construct_params,
                                             GObjectConstructParam  *construct_params);
static void	ibus_kmfl_engine_destroy		(IBusKMFLEngine		    *kmfl);
static gboolean
			ibus_kmfl_engine_process_key_event
                                            (IBusEngine             *engine,
                                             guint               	 keyval,
                                             guint               	 keycode,
                                             guint               	 state);
static void ibus_kmfl_engine_focus_in       (IBusEngine             *engine);
static void ibus_kmfl_engine_focus_out      (IBusEngine             *engine);
static void ibus_kmfl_engine_reset          (IBusEngine             *engine);
static void ibus_kmfl_engine_enable         (IBusEngine             *engine);
static void ibus_kmfl_engine_disable        (IBusEngine             *engine);
static void ibus_engine_set_cursor_location (IBusEngine             *engine,
                                             gint                    x,
                                             gint                    y,
                                             gint                    w,
                                             gint                    h);
static void ibus_kmfl_engine_set_capabilities
                                            (IBusEngine             *engine,
                                             guint                   caps);
static void ibus_kmfl_engine_page_up        (IBusEngine             *engine);
static void ibus_kmfl_engine_page_down      (IBusEngine             *engine);
static void ibus_kmfl_engine_cursor_up      (IBusEngine             *engine);
static void ibus_kmfl_engine_cursor_down    (IBusEngine             *engine);
static void ibus_kmfl_engine_property_activate
                                            (IBusEngine             *engine,
                                             const gchar            *prop_name,
                                             guint                   prop_state);
static void ibus_kmfl_engine_property_show
											(IBusEngine             *engine,
                                             const gchar            *prop_name);
static void ibus_kmfl_engine_property_hide
											(IBusEngine             *engine,
                                             const gchar            *prop_name);

static void ibus_kmfl_engine_commit_string
                                            (IBusKMFLEngine         *kmfl,
                                             const gchar            *string);

static IBusEngineClass *parent_class = NULL;
static GHashTable      *im_table = NULL;


void output_string(void *contrack, char *ptr) {
    KInputContext * ic = (KInputContext *)contrack;
    g_debug("DAR: output_string - ic->cmds=%p", ic->cmds);
    if (ptr) {
        Cmd * cmd = g_new(Cmd, 1);
        cmd->opcode=OUTPUT_STRING;
        cmd->cmdarg=g_strdup(ptr);
        ic->cmds = g_list_append(ic->cmds, cmd);
    }
}

void erase_char(void *contrack) {
    KInputContext * ic = (KInputContext *)contrack;
    g_debug("DAR: erase_char - ic->cmds=%p", ic->cmds);
    Cmd * cmd = g_new(Cmd, 1);
    cmd->opcode = ERASE_CHAR;
    cmd->cmdarg=NULL;
    ic->cmds = g_list_append(ic->cmds, cmd);
}

void output_char(void *contrack, unsigned char byte) {
    if (byte == 8) {
        erase_char(contrack);
    } else {
        char s[2];
        s[0] = byte;
        s[1] = '\0';
        output_string(contrack, s);
    }
}

void forward_keyevent(void *contrack, unsigned int key, unsigned int state)
{
    printf("Not currently supported\n");
}

void output_beep(void *contrack) {
    KInputContext * ic = (KInputContext *)contrack;
    Cmd * cmd = g_new(Cmd, 1);
    cmd->opcode = OUTPUT_BEEP;
    cmd->cmdarg=NULL;
    ic->cmds = g_list_append(ic->cmds, cmd);    
}

/*
 * Create an input context
 */
KInputContext *
kmfl_create_ic(KInputMethod *im)
{
    KInputContext * ic;
    g_debug("DAR: kmfl_create_ic");
    ic = g_new(KInputContext, 1);
    ic->cmds = NULL;
    ic->p_kmsi=kmfl_make_keyboard_instance(ic);
    kmfl_attach_keyboard(ic->p_kmsi, im->keyboard_number);

    return ic;
}


/*
 * Destroy an input context
 */
void
kmfl_destroy_ic(KInputContext *ic)
{
    GList * p;
    g_debug("DAR: kmfl_destroy_ic");
    kmfl_detach_keyboard(ic->p_kmsi);
    kmfl_delete_keyboard_instance(ic->p_kmsi);
    
    if (ic->cmds) {
        for (p=ic->cmds; p != NULL; p = p->next) {
            Cmd * cmd = (Cmd *)p->data;
            if (cmd) {
                if (cmd->cmdarg)
                    g_free(cmd->cmdarg);
                g_free(cmd);
            }
        }
        g_list_free(ic->cmds);
        ic->cmds=NULL;
    }
    
}

GType
ibus_kmfl_engine_get_type (void)
{
	static GType type = 0;

	static const GTypeInfo type_info = {
		sizeof (IBusKMFLEngineClass),
		(GBaseInitFunc)		NULL,
		(GBaseFinalizeFunc) NULL,
		(GClassInitFunc)	ibus_kmfl_engine_class_init,
		NULL,
		NULL,
		sizeof (IBusKMFLEngine),
		0,
		(GInstanceInitFunc)	ibus_kmfl_engine_init,
	};

	if (type == 0) {
		type = g_type_register_static (IBUS_TYPE_ENGINE,
									   "IBusKMFLEngine",
									   &type_info,
									   (GTypeFlags) 0);
	}

	return type;
}

static void
ibus_kmfl_engine_class_init (IBusKMFLEngineClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
	IBusObjectClass *ibus_object_class = IBUS_OBJECT_CLASS (klass);
	IBusEngineClass *engine_class = IBUS_ENGINE_CLASS (klass);

	parent_class = (IBusEngineClass *) g_type_class_peek_parent (klass);

    object_class->constructor = ibus_kmfl_engine_constructor;
	ibus_object_class->destroy = (IBusObjectDestroyFunc) ibus_kmfl_engine_destroy;

    engine_class->process_key_event = ibus_kmfl_engine_process_key_event;

    engine_class->reset = ibus_kmfl_engine_reset;
    engine_class->enable = ibus_kmfl_engine_enable;
    engine_class->disable = ibus_kmfl_engine_disable;

    engine_class->focus_in = ibus_kmfl_engine_focus_in;
    engine_class->focus_out = ibus_kmfl_engine_focus_out;

    engine_class->page_up = ibus_kmfl_engine_page_up;
    engine_class->page_down = ibus_kmfl_engine_page_down;

    engine_class->cursor_up = ibus_kmfl_engine_cursor_up;
    engine_class->cursor_down = ibus_kmfl_engine_cursor_down;

    engine_class->property_activate = ibus_kmfl_engine_property_activate;
}

static void
ibus_kmfl_engine_init (IBusKMFLEngine *kmfl)
{
    kmfl->status_prop = ibus_property_new ("status",
                                           PROP_TYPE_NORMAL,
                                           NULL,
                                           NULL,
                                           NULL,
                                           TRUE,
                                           FALSE,
                                           0,
                                           NULL);

    kmfl->prop_list = ibus_prop_list_new ();
    ibus_prop_list_append (kmfl->prop_list,  kmfl->status_prop);

    kmfl->table = ibus_lookup_table_new (9, 0, TRUE, TRUE);
    kmfl->context = NULL;
}

static GObject*
ibus_kmfl_engine_constructor (GType                   type,
                              guint                   n_construct_params,
                              GObjectConstructParam  *construct_params)
{
    IBusKMFLEngine *kmfl;
    KInputMethod *im;
    const gchar *engine_name;
    
    g_debug("DAR: ibus_kmfl_engine_constructor");
    
    kmfl = (IBusKMFLEngine *) G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_params,
                                                       construct_params);

    engine_name = ibus_engine_get_name ((IBusEngine *) kmfl);
    g_assert (engine_name);

    if (im_table == NULL) {
        im_table = g_hash_table_new_full (g_str_hash,
                                          g_str_equal,
                                          g_free,
                                          (GDestroyNotify) kinput_close_im);
    }

    im = (KInputMethod *) g_hash_table_lookup (im_table, engine_name);

    if (im == NULL) {
        im = kinput_open_im(engine_name);
        
        if (im != NULL)
            g_hash_table_insert (im_table, g_strdup (engine_name), im);

    }

    if (im == NULL) {
        g_warning ("Can not find kmfl keymap %s", engine_name);
        g_object_unref (kmfl);
        return NULL;
    }

    kmfl->context = kmfl_create_ic (im);

    return (GObject *) kmfl;
}


static void
ibus_kmfl_engine_destroy (IBusKMFLEngine *kmfl)
{
    const gchar *engine_name;
    g_debug("DAR: ibus_kmfl_engine_destroy");
    
    engine_name = ibus_engine_get_name ((IBusEngine *) kmfl);
    g_assert (engine_name);

    if (kmfl->prop_list) {
        g_object_unref (kmfl->prop_list);
        kmfl->prop_list = NULL;
    }

    if (kmfl->status_prop) {
        g_object_unref (kmfl->status_prop);
        kmfl->status_prop = NULL;
    }

    if (kmfl->table) {
        g_object_unref (kmfl->table);
        kmfl->table = NULL;
    }

    if (kmfl->context) {
        kmfl_destroy_ic (kmfl->context);
        kmfl->context = NULL;
    }
    
    g_hash_table_remove(im_table, engine_name);
     
	IBUS_OBJECT_CLASS (parent_class)->destroy ((IBusObject *)kmfl);
}

static void
ibus_kmfl_engine_commit_string (IBusKMFLEngine *kmfl,
                                const gchar    *string)
{
    IBusText *text;
    text = ibus_text_new_from_static_string (string);
    ibus_engine_commit_text ((IBusEngine *)kmfl, text);
    g_object_unref (text);
}

static gboolean
ibus_kmfl_engine_process_key_event (IBusEngine     *engine,
                                    guint           keyval,
                                    guint           keycode,
                                    guint           state)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;
    
    if (state & IBUS_RELEASE_MASK)
        return FALSE;

    g_debug("DAR: ibus_kmfl_engine_process_key_event - keyval=%x, state=%x", keyval, state);

    if (kmfl_interpret(kmfl->context->p_kmsi, keyval, state) == 1) {
        GList * p;
        for (p=kmfl->context->cmds; p != NULL; p = p->next) {
            Cmd * cmd = (Cmd *) p->data;
            if (cmd) {
                switch (cmd->opcode) {
                    case OUTPUT_STRING:
                        ibus_kmfl_engine_commit_string(kmfl, cmd->cmdarg);
                        break;
                    case ERASE_CHAR:
                        if (((engine->client_capabilities & IBUS_CAP_SURROUNDING_TEXT) != 0) && (p->next != NULL))
                            ibus_engine_delete_surrounding_text(engine, -1, 1);
                        else 
                            ibus_engine_forward_key_event(engine, IBUS_BackSpace, 22, 0);
                        break;
                    case FORWARD_KEYEVENT:
                        break;
                    case OUTPUT_BEEP:
                        break;
                }
                if (cmd->cmdarg)
                    g_free(cmd->cmdarg);
                g_free(cmd);
            }
        }
        g_list_free(kmfl->context->cmds);
        kmfl->context->cmds=NULL;
        return TRUE;
    }
    
    return FALSE;
 }

static void
ibus_kmfl_engine_focus_in (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    ibus_engine_register_properties (engine, kmfl->prop_list);

    parent_class->focus_in (engine);
}

static void
ibus_kmfl_engine_focus_out (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->focus_out (engine);
}

static void
ibus_kmfl_engine_reset (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->reset (engine);
    ibus_kmfl_engine_focus_in (engine);
}

static void
ibus_kmfl_engine_enable (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->enable (engine);
}

static void
ibus_kmfl_engine_disable (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    ibus_kmfl_engine_focus_out (engine);
    parent_class->disable (engine);
}

static void
ibus_kmfl_engine_page_up (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->page_up (engine);
}

static void
ibus_kmfl_engine_page_down (IBusEngine *engine)
{

    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->page_down (engine);
}

static void
ibus_kmfl_engine_cursor_up (IBusEngine *engine)
{

    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->cursor_up (engine);
}

static void
ibus_kmfl_engine_cursor_down (IBusEngine *engine)
{

    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->cursor_down (engine);
}

static void
ibus_kmfl_engine_property_activate (IBusEngine  *engine,
                                    const gchar *prop_name,
                                    guint        prop_state)
{
    parent_class->property_activate (engine, prop_name, prop_state);
}

==> ibus-kmfl-1.0.3.tar.gz: ibus-kmfl-1.0.3/src/engine.c <==
/* vim:set et sts=4: */

#include <ibus.h>
#include <string.h>
#include <stdio.h>
#include "kmflutil.h"
#include "engine.h"

typedef struct _IBusKMFLEngine IBusKMFLEngine;
typedef struct _IBusKMFLEngineClass IBusKMFLEngineClass;

struct _IBusKMFLEngine {
	IBusEngine parent;

    /* members */
    KInputContext *context;
    IBusLookupTable *table;
    IBusProperty    *status_prop;
    IBusPropList    *prop_list;
};

struct _IBusKMFLEngineClass {
	IBusEngineClass parent;
};

/* functions prototype */
static void	ibus_kmfl_engine_class_init	    (IBusKMFLEngineClass    *klass);
static void	ibus_kmfl_engine_init		    (IBusKMFLEngine		    *kmfl);
static GObject*
            ibus_kmfl_engine_constructor    (GType                   type,
                                             guint                   n_construct_params,
                                             GObjectConstructParam  *construct_params);
static void	ibus_kmfl_engine_destroy		(IBusKMFLEngine		    *kmfl);
static gboolean
			ibus_kmfl_engine_process_key_event
                                            (IBusEngine             *engine,
                                             guint               	 keyval,
                                             guint               	 keycode,
                                             guint               	 state);
static void ibus_kmfl_engine_focus_in       (IBusEngine             *engine);
static void ibus_kmfl_engine_focus_out      (IBusEngine             *engine);
static void ibus_kmfl_engine_reset          (IBusEngine             *engine);
static void ibus_kmfl_engine_enable         (IBusEngine             *engine);
static void ibus_kmfl_engine_disable        (IBusEngine             *engine);
static void ibus_engine_set_cursor_location (IBusEngine             *engine,
                                             gint                    x,
                                             gint                    y,
                                             gint                    w,
                                             gint                    h);
static void ibus_kmfl_engine_set_capabilities
                                            (IBusEngine             *engine,
                                             guint                   caps);
static void ibus_kmfl_engine_page_up        (IBusEngine             *engine);
static void ibus_kmfl_engine_page_down      (IBusEngine             *engine);
static void ibus_kmfl_engine_cursor_up      (IBusEngine             *engine);
static void ibus_kmfl_engine_cursor_down    (IBusEngine             *engine);
static void ibus_kmfl_engine_property_activate
                                            (IBusEngine             *engine,
                                             const gchar            *prop_name,
                                             guint                   prop_state);
static void ibus_kmfl_engine_property_show
											(IBusEngine             *engine,
                                             const gchar            *prop_name);
static void ibus_kmfl_engine_property_hide
											(IBusEngine             *engine,
                                             const gchar            *prop_name);

static void ibus_kmfl_engine_commit_string
                                            (IBusKMFLEngine         *kmfl,
                                             const gchar            *string);

static IBusEngineClass *parent_class = NULL;
static GHashTable      *im_table = NULL;


void output_string(void *contrack, char *ptr) {
    KInputContext * ic = (KInputContext *)contrack;
    g_debug("DAR: output_string - ic->cmds=%p", ic->cmds);
    if (ptr) {
        Cmd * cmd = g_new(Cmd, 1);
        cmd->opcode=OUTPUT_STRING;
        cmd->cmdarg=g_strdup(ptr);
        ic->cmds = g_list_append(ic->cmds, cmd);
    }
}

void erase_char(void *contrack) {
    KInputContext * ic = (KInputContext *)contrack;
    g_debug("DAR: erase_char - ic->cmds=%p", ic->cmds);
    Cmd * cmd = g_new(Cmd, 1);
    cmd->opcode = ERASE_CHAR;
    cmd->cmdarg=NULL;
    ic->cmds = g_list_append(ic->cmds, cmd);
}

void output_char(void *contrack, unsigned char byte) {
    if (byte == 8) {
        erase_char(contrack);
    } else {
        char s[2];
        s[0] = byte;
        s[1] = '\0';
        output_string(contrack, s);
    }
}

void forward_keyevent(void *contrack, unsigned int key, unsigned int state)
{
    printf("Not currently supported\n");
}

void output_beep(void *contrack) {
    KInputContext * ic = (KInputContext *)contrack;
    Cmd * cmd = g_new(Cmd, 1);
    cmd->opcode = OUTPUT_BEEP;
    cmd->cmdarg=NULL;
    ic->cmds = g_list_append(ic->cmds, cmd);    
}

/*
 * Create an input context
 */
KInputContext *
kmfl_create_ic(KInputMethod *im)
{
    KInputContext * ic;
    g_debug("DAR: kmfl_create_ic");
    ic = g_new(KInputContext, 1);
    ic->cmds = NULL;
    ic->p_kmsi=kmfl_make_keyboard_instance(ic);
    kmfl_attach_keyboard(ic->p_kmsi, im->keyboard_number);

    return ic;
}


/*
 * Destroy an input context
 */
void
kmfl_destroy_ic(KInputContext *ic)
{
    GList * p;
    g_debug("DAR: kmfl_destroy_ic");
    kmfl_detach_keyboard(ic->p_kmsi);
    kmfl_delete_keyboard_instance(ic->p_kmsi);
    
    if (ic->cmds) {
        for (p=ic->cmds; p != NULL; p = p->next) {
            Cmd * cmd = (Cmd *)p->data;
            if (cmd) {
                if (cmd->cmdarg)
                    g_free(cmd->cmdarg);
                g_free(cmd);
            }
        }
        g_list_free(ic->cmds);
        ic->cmds=NULL;
    }
    
}

GType
ibus_kmfl_engine_get_type (void)
{
	static GType type = 0;

	static const GTypeInfo type_info = {
		sizeof (IBusKMFLEngineClass),
		(GBaseInitFunc)		NULL,
		(GBaseFinalizeFunc) NULL,
		(GClassInitFunc)	ibus_kmfl_engine_class_init,
		NULL,
		NULL,
		sizeof (IBusKMFLEngine),
		0,
		(GInstanceInitFunc)	ibus_kmfl_engine_init,
	};

	if (type == 0) {
		type = g_type_register_static (IBUS_TYPE_ENGINE,
									   "IBusKMFLEngine",
									   &type_info,
									   (GTypeFlags) 0);
	}

	return type;
}

static void
ibus_kmfl_engine_class_init (IBusKMFLEngineClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);
	IBusObjectClass *ibus_object_class = IBUS_OBJECT_CLASS (klass);
	IBusEngineClass *engine_class = IBUS_ENGINE_CLASS (klass);

	parent_class = (IBusEngineClass *) g_type_class_peek_parent (klass);

    object_class->constructor = ibus_kmfl_engine_constructor;
	ibus_object_class->destroy = (IBusObjectDestroyFunc) ibus_kmfl_engine_destroy;

    engine_class->process_key_event = ibus_kmfl_engine_process_key_event;

    engine_class->reset = ibus_kmfl_engine_reset;
    engine_class->enable = ibus_kmfl_engine_enable;
    engine_class->disable = ibus_kmfl_engine_disable;

    engine_class->focus_in = ibus_kmfl_engine_focus_in;
    engine_class->focus_out = ibus_kmfl_engine_focus_out;

    engine_class->page_up = ibus_kmfl_engine_page_up;
    engine_class->page_down = ibus_kmfl_engine_page_down;

    engine_class->cursor_up = ibus_kmfl_engine_cursor_up;
    engine_class->cursor_down = ibus_kmfl_engine_cursor_down;

    engine_class->property_activate = ibus_kmfl_engine_property_activate;
}

static void
ibus_kmfl_engine_init (IBusKMFLEngine *kmfl)
{
    kmfl->status_prop = ibus_property_new ("status",
                                           PROP_TYPE_NORMAL,
                                           NULL,
                                           NULL,
                                           NULL,
                                           TRUE,
                                           FALSE,
                                           0,
                                           NULL);

    kmfl->prop_list = ibus_prop_list_new ();
    ibus_prop_list_append (kmfl->prop_list,  kmfl->status_prop);

    kmfl->table = ibus_lookup_table_new (9, 0, TRUE, TRUE);
    kmfl->context = NULL;
}

static GObject*
ibus_kmfl_engine_constructor (GType                   type,
                              guint                   n_construct_params,
                              GObjectConstructParam  *construct_params)
{
    IBusKMFLEngine *kmfl;
    KInputMethod *im;
    const gchar *engine_name;
    
    g_debug("DAR: ibus_kmfl_engine_constructor");
    
    kmfl = (IBusKMFLEngine *) G_OBJECT_CLASS (parent_class)->constructor (type,
                                                       n_construct_params,
                                                       construct_params);

    engine_name = ibus_engine_get_name ((IBusEngine *) kmfl);
    g_assert (engine_name);

    if (im_table == NULL) {
        im_table = g_hash_table_new_full (g_str_hash,
                                          g_str_equal,
                                          g_free,
                                          (GDestroyNotify) kinput_close_im);
    }

    im = (KInputMethod *) g_hash_table_lookup (im_table, engine_name);

    if (im == NULL) {
        im = kinput_open_im(engine_name);
        
        if (im != NULL)
            g_hash_table_insert (im_table, g_strdup (engine_name), im);

    }

    if (im == NULL) {
        g_warning ("Can not find kmfl keymap %s", engine_name);
        g_object_unref (kmfl);
        return NULL;
    }

    kmfl->context = kmfl_create_ic (im);

    return (GObject *) kmfl;
}


static void
ibus_kmfl_engine_destroy (IBusKMFLEngine *kmfl)
{
    const gchar *engine_name;
    g_debug("DAR: ibus_kmfl_engine_destroy");
    
    engine_name = ibus_engine_get_name ((IBusEngine *) kmfl);
    g_assert (engine_name);

    if (kmfl->prop_list) {
        g_object_unref (kmfl->prop_list);
        kmfl->prop_list = NULL;
    }

    if (kmfl->status_prop) {
        g_object_unref (kmfl->status_prop);
        kmfl->status_prop = NULL;
    }

    if (kmfl->table) {
        g_object_unref (kmfl->table);
        kmfl->table = NULL;
    }

    if (kmfl->context) {
        kmfl_destroy_ic (kmfl->context);
        kmfl->context = NULL;
    }
    
    g_hash_table_remove(im_table, engine_name);
     
	IBUS_OBJECT_CLASS (parent_class)->destroy ((IBusObject *)kmfl);
}

static void
ibus_kmfl_engine_commit_string (IBusKMFLEngine *kmfl,
                                const gchar    *string)
{
    IBusText *text;
    text = ibus_text_new_from_static_string (string);
    ibus_engine_commit_text ((IBusEngine *)kmfl, text);
    g_object_unref (text);
}

static gboolean
ibus_kmfl_engine_process_key_event (IBusEngine     *engine,
                                    guint           keyval,
                                    guint           keycode,
                                    guint           state)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;
    
    if (state & IBUS_RELEASE_MASK)
        return FALSE;

    g_debug("DAR: ibus_kmfl_engine_process_key_event - keyval=%x, keycode=%x, state=%x", keyval, keycode, state);

    // Let the application handle user generated backspaces after resetting the kmfl history
    if (keyval == IBUS_BackSpace) {
        clear_history(kmfl->context->p_kmsi);
    } else if (kmfl_interpret(kmfl->context->p_kmsi, keyval, state) == 1) {
        GList * p;
        for (p=kmfl->context->cmds; p != NULL; p = p->next) {
            Cmd * cmd = (Cmd *) p->data;
            if (cmd) {
                switch (cmd->opcode) {
                    case OUTPUT_STRING:
                        ibus_kmfl_engine_commit_string(kmfl, cmd->cmdarg);
                        break;
                    case ERASE_CHAR:
                        if ((engine->client_capabilities & IBUS_CAP_SURROUNDING_TEXT) != 0)
                            ibus_engine_delete_surrounding_text(engine, -1, 1);
                        else 
                            ibus_engine_forward_key_event(engine, IBUS_BackSpace, 22, 0);
                        break;
                    case FORWARD_KEYEVENT:
                        break;
                    case OUTPUT_BEEP:
                        break;
                }
                if (cmd->cmdarg)
                    g_free(cmd->cmdarg);
                g_free(cmd);
            }
        }
        g_list_free(kmfl->context->cmds);
        kmfl->context->cmds=NULL;
        return TRUE;
    }
    
    return FALSE;
 }

static void
ibus_kmfl_engine_focus_in (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    ibus_engine_register_properties (engine, kmfl->prop_list);

    parent_class->focus_in (engine);
}

static void
ibus_kmfl_engine_focus_out (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->focus_out (engine);
}

static void
ibus_kmfl_engine_reset (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->reset (engine);
    ibus_kmfl_engine_focus_in (engine);
}

static void
ibus_kmfl_engine_enable (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->enable (engine);
}

static void
ibus_kmfl_engine_disable (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    ibus_kmfl_engine_focus_out (engine);
    parent_class->disable (engine);
}

static void
ibus_kmfl_engine_page_up (IBusEngine *engine)
{
    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->page_up (engine);
}

static void
ibus_kmfl_engine_page_down (IBusEngine *engine)
{

    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->page_down (engine);
}

static void
ibus_kmfl_engine_cursor_up (IBusEngine *engine)
{

    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->cursor_up (engine);
}

static void
ibus_kmfl_engine_cursor_down (IBusEngine *engine)
{

    IBusKMFLEngine *kmfl = (IBusKMFLEngine *) engine;

    parent_class->cursor_down (engine);
}

static void
ibus_kmfl_engine_property_activate (IBusEngine  *engine,
                                    const gchar *prop_name,
                                    guint        prop_state)
{
    parent_class->property_activate (engine, prop_name, prop_state);
}
